/*
 * LOD demo
 *
 * Copyright (c) 2002 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include "precompiled.h"

#include "wposix.h"
#include "wsdl.h"
#include "font.h"
#include "lod.h"
#include "system.h"
#include "util.h"
#include "input.h"
#include "res.h"
#include "mpd.h"



/* viewer pos, velocity and orientation */
float x = 0.f, y = 0.f, z = 150.f;
float v = 0.f, d_v = 0.f;
float pitch = 0.f, d_pitch = 0.f;
float heading = 0.f, d_heading = 0.f;

uint32 game_ticks = 0;

extern int fps;

uint fd;

/* video mode */
int w = 1024, h = 768, bpp = 16, fullscreen = 1;


static void init_terrain()
{
	if(lod_load_preproc_data("mpd_2k.pt") < 0)
	{
		int size = 2049;
		uchar* phmap = gen(size);
		lod_preproc(size, phmap);
		free(phmap);
		mkdir("data", 0777);
		lod_save_preproc_data("data\\mpd_2k.pt");
	}
}


static void render(void)
{
	glClear(GL_COLOR_BUFFER_BIT);

	/* move viewpoint */
	glLoadIdentity();
	gluLookAt(x,y,z, x+cos(heading)*cos(pitch),y+sin(heading)*cos(pitch),z+sin(pitch), 0.,0.,1.);
	lod_latch_view_pos();

	lod_render();

	/* fps counter */
	glPushAttrib(GL_ENABLE_BIT);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	glEnable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);
	glBlendFunc(GL_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);
	glAlphaFunc(GL_GREATER, 0.5);
	glEnable(GL_ALPHA_TEST);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(10, 30, 0);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0., w, 0., h, -1., 1.);
	bind_font(fd);
	glprintf("%d fps", fps);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glPopAttrib();
}


static bool input_handler(const SDL_Event* event)
{
	switch(event->type)
	{
	case SDL_KEYDOWN:
		switch(event->key.keysym.sym)
		{
			case SDLK_8: d_v = v = 0; break;
			case SDLK_9: d_v = -.1f; break;
			case SDLK_0: d_v =  .1f; break;
			case SDLK_ESCAPE: exit(0);
		}
		break;

	case SDL_KEYUP:
		switch(event->key.keysym.sym)
		{
			case SDLK_8: d_v = 0; break;
			case SDLK_9: if(d_v < 0) d_v = 0; break;
			case SDLK_0: if(d_v > 0) d_v = 0; break;

			case SDLK_r: in_record("rec");break;
			case SDLK_s: in_stop();
			case SDLK_p: in_playback("rec");break;
		}
		break;

	case SDL_MOUSEMOTION:
		d_heading = ((w/2)-event->motion.x)*0.0005f;
		d_pitch   = ((h/2)-event->motion.y)*0.0005f*4.f/3.f;
		break;
	}

	return true;
}


static void do_tick()
{
	heading += d_heading;
	pitch += d_pitch;
	v += d_v;

	x += cosf(heading)*cosf(pitch)*v; y += sinf(heading)*cosf(pitch)*v; z += sinf(pitch)*v;
}


int main(int /* argc */, char* /* argv[] */)
{
	if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_AUDIO|SDL_INIT_TIMER|SDL_INIT_NOPARACHUTE) < 0)
	{
		printf("Problem initializing SDL: %.100s\n", SDL_GetError());
		return 1;
	}
	atexit(SDL_Quit);

	if(set_video_mode(w, h, bpp, fullscreen) < 0)
		return -1;

	sys_init();
#ifdef x86
	printf("%s, %d MHz\n", sys_cpu_type, (int)(sys_cpu_freq*1e-6));
#endif
	printf("%s %s %s\n", glGetString(GL_VENDOR), glGetString(GL_RENDERER), glGetString(GL_VERSION));

	in_add_handler(input_handler);

	res_add_search_path("data");

	fd = load_font("verdana.fnt");

	init_terrain();

	/* disable vsync */
	if(ext_avail("WGL_EXT_swap_control"))
		wglSwapIntervalEXT(0);

	/* gl state */
	glFrontFace(GL_CW);
	glEnable(GL_CULL_FACE);
	glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60., 4./3., 1., 16000.);
	glMatrixMode(GL_MODELVIEW);

	SDL_WarpMouse(w/2, h/2);
	SDL_WM_GrabInput(SDL_GRAB_ON);


//in_playback("rec");


/* fixed timestep main loop */

#define TICK_TIME 30

	uint32 time0 = SDL_GetTicks();
	while(1)
	{
		uint32 time1 = SDL_GetTicks();
		while((time1-time0) > TICK_TIME)
		{
			game_ticks++;
			in_get_events();
			do_tick();
			time0 += TICK_TIME;
		}

 		render();
		SDL_GL_SwapBuffers();

		calc_fps();
	}

	return 0;	/* ANSI compliance */
}