/*
 * quick'n dirty midpoint displacement terrain generator
 * with unidirectional FIR filter
 *
 * Copyright (c) 2002 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include "precompiled.h"


#define displacement	120.f
#define roughness		0.7f


/*
 * generate a pseudo-random float with abs value <= displacement/2
 * "random algorithm" taken from VC++ CRT's rand()
 */

static float r_scale;

#define RAND_MAX 0x7fff
static long seed = 1;

static inline float r()
{
	seed *= 0x000343fd; seed += 0x00269ec3;

	return (((seed >> 16) & RAND_MAX) - RAND_MAX/2) * r_scale;
}


uchar* gen(uint w)
{
	uint i, x, y;
	uint step = w-1;
	uint cnt = 1;

	float *p, *p2;
	float s, t;

#define h(x,y) _h[(y)*w+(x)]
	float* _h = (float*)calloc(w*w, sizeof(float));

	r_scale = displacement / RAND_MAX;

	do
	{
		int dx = step * 4, dy = step * w * 4;
		int dx2 = dx/2, dy2 = dy/2;
		int dsw2 = dy2-dx2, dse2 = dx2+dy2, dne2 = dx2-dy2;

/* return ptr to float <ofs> bytes behind p */
#define fp(p, ofs) ((float*)(((char*)(p))+(ofs)))

		/* diamond */
		for(y = 0; y < w-1; y += step)
		{
			p = &h(0, y);
			s = *p + *fp(p, dy);
			x = cnt;
			do
			{
				t = s;
				p = fp(p, dx);	/* p += step */
				s = *p + *fp(p, dy);
				*fp(p, dsw2) = (t + s) * 0.25f + r();
			}
			while(--x);
		}

		/* square */
		for(y = step; y < w-1; y += step)
		{
			p = &h(step, y);
			s = *p + *fp(p, dsw2);
			x = cnt-1;
			do
			{
				t = *fp(p, dse2);
				*fp(p, dy2) = (s + t + *fp(p, dy)) * 0.25f + r();
				p2 = fp(p, dx);	/* p2 = p + step */
				s = *p2 + t;
				*fp(p, dx2) = (s + *p + *fp(p, dne2)) * 0.25f + r();
				p = p2;
			}
			while(--x);
		}

		cnt += cnt;
		step >>= 1;
		r_scale *= powf(2, -roughness);
	}
	while(cnt < w-1);


#define a 0.7f
#define b 0.3f

	/* forward fir filter (rows) */
	for(y = 0; y < w; y++)
	{
		p = &h(0, y);
		i = w-1;
		do
		{
			t = *p++ * a;
			*p *= b;
			*p += t;
		}
		while(--i);
	}

	/* forward fir filter (columns) */
	for(y = 1; y < w; y++)
	{
		p = &h(0, y-1);
		p2 = &h(0, y);
		i = w-1;
		do
		{
			t = *p++ * a;
			*p2 *= b;
			*p2 += t;
			p2++;
		}
		while(--i);
	}

	/* calculate min, max, range (used to normalize height values) */
	float min = 1e6, max = 0.f;
	p = _h;
	for(i = 0; i < w*w; i++)
	{
		t = *p++;
		max = t > max? t : max;
		min = t < min? t : min;
	}
	float scale = 1.f / (max - min);

	/* normalize, square (for smoothing), and convert to uchar (in place) */
	p = _h;
	uchar* c = (uchar*)_h;
	for(i = 0; i < w*w; i++)
	{
		t = (*p++ - min) * scale;
		*c++ = (uchar)ceil(t * t * 255.f);
	}

	return (uchar*)realloc(_h, w*w*1);
}