/*
 * resource layer
 * loads files from ZIP archives / disk transparently,
 * with multiple search paths
 *
 * Copyright (c) 2002 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include "precompiled.h"

#include "wposix.h"

#include "res.h"
#include "system.h"
#include "unzip.h"


typedef struct PATH
{
	int zd;
	char path[SYS_MAX_PATH];
	struct PATH* next;
}
PATH;
static PATH* path_list;


enum
{
	MAPPED,
	ZIP,
	MEM
};

typedef struct
{
	int type;
	void* ptr;
	int size;
}
RES;
static RES* res_array;
static uint res_size, res_entries;

/* helper function for res_load(), res_malloc() */
static void add_res_entry(int type, void* ptr, int size)
{
	if(res_entries >= res_size)
	{
		res_size += 10;
		res_array = (RES*)realloc(res_array, res_size*sizeof(RES));
		memset(&res_array[res_entries], 0, 10*sizeof(RES));
	}

	RES* res = &res_array[res_entries++];
	res->type = type;
	res->ptr = ptr;
	res->size = size;
}


typedef struct
{
	int width, height;
	GLenum fmt;
	uchar* ptr;
}
TEX_INFO;
static TEX_INFO* ti_array;
static uint ti_size;


static void cleanup(void)
{
	uint i;

	/* free path linked list */
	PATH *entry = path_list, *temp;
	while(entry)
	{
		temp = entry;
		entry = entry->next;
		free(temp);
	}

	/* close all open resources */
	for(i = 0; i < res_size; i++)
		res_free(res_array[i].ptr);
	free(res_array);

	free(ti_array);	/* images already freed by loop above */
}


int res_add_search_path(const char* path)
{
	/* first call: setup cleanup atexit routine */
	if(!path_list)
		atexit(cleanup);

	int valid = 0;

	/* check if path is a dir */
	DIR* dir = opendir(path);
	if(dir)
	{
		valid = 1;
		closedir(dir);
	}

	/* check if path is a zip file */
	char zip_fn[SYS_MAX_PATH];
	sprintf(zip_fn, "%s.zip", path);
	int zd = open_zip(zip_fn);
	if(zd >= 0)
		valid = 1;

	/* neither: abort */
	if(!valid)
		return -1;

	/* add entry to front of path list */
	PATH* entry = (PATH*)calloc(1, sizeof(PATH));
	entry->next = path_list;
	path_list = entry;

	/* fill entry fields */
	entry->zd = zd;
	strncpy(entry->path, path, sizeof(entry->path));

	return 0;
}


void* res_load(const char* fn, int* _size)
{
	static char path[SYS_MAX_PATH];
	struct stat stat_buf;

	int type;
	void* ptr;
	int size;

	if(_size)
		*_size = 0;

	for(PATH* entry = path_list; entry; entry = entry->next)
	{
		/* dir */
		sprintf(path, "%s%c%s", entry->path, SYS_DIR_SEP, fn);
		int fd = open(path, O_RDONLY);
		if(fd >= 0 && stat(path, &stat_buf) == 0)
		{
			size = stat_buf.st_size;
			ptr = mmap(0, size, PROT_READ, MAP_PRIVATE, fd, 0);
			close(fd);
			if(ptr != MAP_FAILED)
			{
				type = MAPPED;
				goto found_it;
			}
		}

		/* zip */
		ptr = unzip(entry->zd, fn, &size);
		if(ptr)
		{
			type = ZIP;
			goto found_it;
		}
	}

	return NULL;

found_it:
	add_res_entry(type, ptr, size);

	if(_size)
		*_size = size;

	return ptr;
}


void* res_malloc(const int size)
{
	void* ptr = malloc(size);
	if(!ptr)
		return NULL;

	add_res_entry(MEM, ptr, size);

	return ptr;
}


void res_free(void* ptr)
{
	if(!ptr)
		return;

	RES* res = res_array;
	for(uint i = 0; i < res_size; i++, res++)
		if(res->ptr == ptr)
		{
			if(res->type == MAPPED)
				munmap(res->ptr, res->size);
			else
				free(res->ptr);

			res->ptr = NULL;
			break;
		}
}


uint load_texture(const char* fn)
{
	int size;

	/* load image */
	uchar* ptr = (uchar*)res_load(fn, &size);
	if(!ptr)
		return 0;

	/* alloc */
	uint id;
	glGenTextures(1, &id);
	if(id >= ti_size)
	{
		ti_size = id*2;
		ti_array = (TEX_INFO*)realloc(ti_array, ti_size*sizeof(TEX_INFO));
		memset(&ti_array[id], 0, (ti_size-id)*sizeof(TEX_INFO));
	}

	/*
	 * guess texture dimensions / format
	 * width = height = 2^x
	 */
	static GLenum fmt[5] = { 0, 0, GL_LUMINANCE_ALPHA, GL_RGB, GL_RGBA };
	for(int i = 2; i <= 4; i++)
	{
		int dim = (int)sqrt(size/i);
		if(dim % i || (dim & (dim-1)) || dim*dim*i != size)
			continue;

		TEX_INFO* ti = &ti_array[id];
		ti->fmt = fmt[i];
		ti->width = ti->height = dim;
		ti->ptr = ptr;

		return id;	/* success */
	}

	return 0;	/* invalid texture dimensions */
}


int upload_texture(const uint id, const int quality)
{
	const TEX_INFO* ti = &ti_array[id];

	/* verify width, height == 2^n */
	if(ti->width != ti->height || ti->width&(ti->width-1))
		return 0;

	glBindTexture(GL_TEXTURE_2D, id);

	/* set texture filter params based on quality */
	switch(quality)
	{
	case 0:
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexImage2D(GL_TEXTURE_2D, 0, ti->fmt, ti->width, ti->height, 0, ti->fmt, GL_UNSIGNED_BYTE, ti->ptr);
		break;
	}

	return 0;
}