/*
 * CPU init / automatic OpenGL extension loading
 *
 * Copyright (c) 2002 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include "precompiled.h"

#include "wsdl.h"

#include "system.h"


/************************************************************************************************
 * OpenGL extensions
 ************************************************************************************************/

PFNGLACTIVETEXTUREARBPROC glActiveTextureARB;

/* GL_NV_vertex_array_range */
PFNGLVERTEXARRAYRANGENVPROC glVertexArrayRangeNV;
PFNGLFLUSHVERTEXARRAYRANGENVPROC glFlushVertexArrayRangeNV;

PFNWGLALLOCATEMEMORYNVPROC wglAllocateMemoryNV;
PFNWGLFREEMEMORYNVPROC wglFreeMemoryNV;

/* GL_NV_fence */
PFNGLGENFENCESNVPROC glGenFencesNV;
PFNGLDELETEFENCESNVPROC glDeleteFencesNV;
PFNGLSETFENCENVPROC glSetFenceNV;
PFNGLTESTFENCENVPROC glTestFenceNV;

/* WGL_EXT_swap_control */
PFNWGLSWAPINTERVALEXTPROC wglSwapIntervalEXT;


/* check if the extension <ext> is supported by the OpenGL implementation */
int ext_avail(const char* ext)
{
	char* exts = (char*)glGetString(GL_EXTENSIONS);

	char *p = exts, *end;

	/* make sure ext is valid & doesn't contain spaces */
	if(!ext || ext == '\0' || strchr(ext, ' '))
		return 0;

	while(1)
	{
		p = strstr(p, ext);
		if(!p)
			return 0; /* <ext> string not found - extension not supported */
		end = p + strlen(ext); /* end of current substring */

		/*
		 * make sure the substring found is an entire extension string,
		 * i.e. it starts and ends with ' '
		 */
		if(p == exts || *(p-1) == ' ') /* valid start? */
			if(*end == ' ' || *end == '\0') /* valid end? */
				return 1;
		p = end;
	}
}


/*
 * get function pointers
 *
 * called by set_video_mode(), since pointers may change with pixel format
 */
static void init_gl_ext()
{
	/* GL_ARB_multitexture */
	glActiveTextureARB = (PFNGLACTIVETEXTUREARBPROC)SDL_GL_GetProcAddress("glActiveTextureARB");

	/* GL_NV_vertex_array_range */
	glVertexArrayRangeNV = (PFNGLVERTEXARRAYRANGENVPROC)SDL_GL_GetProcAddress("glVertexArrayRangeNV");
	glFlushVertexArrayRangeNV = (PFNGLFLUSHVERTEXARRAYRANGENVPROC)SDL_GL_GetProcAddress("glFlushVertexArrayRangeNV");
	wglAllocateMemoryNV = (PFNWGLALLOCATEMEMORYNVPROC)SDL_GL_GetProcAddress("wglAllocateMemoryNV");
	wglFreeMemoryNV = (PFNWGLFREEMEMORYNVPROC)SDL_GL_GetProcAddress("wglFreeMemoryNV");

	/* GL_NV_fence */
	glGenFencesNV = (PFNGLGENFENCESNVPROC)SDL_GL_GetProcAddress("glGenFencesNV");
	glDeleteFencesNV = (PFNGLDELETEFENCESNVPROC)SDL_GL_GetProcAddress("glDeleteFencesNV");
	glSetFenceNV = (PFNGLSETFENCENVPROC)SDL_GL_GetProcAddress("glSetFenceNV");
	glTestFenceNV = (PFNGLTESTFENCENVPROC)SDL_GL_GetProcAddress("glTestFenceNV");

	/* WGL_EXT_swap_control */
	wglSwapIntervalEXT = (PFNWGLSWAPINTERVALEXTPROC)SDL_GL_GetProcAddress("wglSwapIntervalEXT");
}


/************************************************************************************************
 * misc
 ************************************************************************************************/

char sys_error_str[256];


int set_video_mode(int w, int h, int bpp, int fullscreen, int depth_size)
{
    SDL_GL_SetAttribute(SDL_GL_DEPTH_SIZE, depth_size);
    SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);

	int flags = SDL_OPENGL;
	if(fullscreen)
		flags |= SDL_FULLSCREEN;

	if(!SDL_SetVideoMode(w, h, bpp, flags))
	{
		sprintf(sys_error_str, "Could not set \"%dx%d:%d, depth %d\" %svideo mode", w, h, bpp, depth_size, fullscreen? "fullscreen ":"");
		return -1;
	}

	glViewport(0, 0, w, h);

	init_gl_ext();

	return 0;
}


/************************************************************************************************
 * cpu
 ************************************************************************************************/

#ifdef x86

char sys_cpu_type[48] = "unknown CPU";
float sys_cpu_freq = 0.f;

bool sys_has_3dnow_pro = false;


static void get_x86_cpu_info()
{
	/*
	 * manual detect for older CPUs
	 */

	char cpu_vendor[13];
	int family, model;

	__asm
	{
		; get vendor string
		xor			eax, eax
		cpuid
		mov			dword ptr [cpu_vendor], ebx
		mov			dword ptr [cpu_vendor+4], edx
		mov			dword ptr [cpu_vendor+8], ecx
		mov			[cpu_vendor+12], 0		; zero terminate string

		; get CPU signature
		mov			eax, 1
		cpuid
		mov			edx, eax
		shr			edx, 4
		and			edx, 0x0f
		mov			[model], edx
		shr			eax, 8
		and			eax, 0x0f
		mov			[family], eax
	}

	/* AMD */
	if(!strcmp(cpu_vendor, "AuthenticAMD"))
	{
		if(family == 6)
			strcpy(sys_cpu_type, (model == 3)? "AMD Duron" : "AMD Athlon");
	}
	/* Intel */
	else if(!strcmp(cpu_vendor, "GenuineIntel"))
	{
		if(family == 6 && model >= 7)
			strcpy(sys_cpu_type, "Intel Pentium III / Celeron");
	}


	/*
	 * check CPU brand string / 3DNow! support on newer CPUs
	 */

	__asm
	{
		; make sure CPUID extended functions are supported
		mov			eax, 0x80000000
		cpuid
		cmp			eax, 0x80000000
		jbe			$no_ext_funct

		; get CPU brand string
		mov			edi, offset sys_cpu_type
		mov			esi, -2					; loop counter: -2 to 0
	$1:	lea			eax, [0x80000004+esi]
		cpuid
		mov			[edi+0], eax
		mov			[edi+4], ebx
		mov			[edi+8], ecx
		mov			[edi+12], edx
		add			edi, 16
		inc			esi
		jle			$1

		; check for 3DNow! Professional
		mov			eax, 0x80000001
		cpuid
		and			edx, 0xc0400000			; 3DNow! & 3DNow! Pro & MMX Ext
		cmp			edx, 0xc0400000
		sete		[sys_has_3dnow_pro]
	$no_ext_funct:
	}


	/*
	 * calculate CPU clock frequency
	 * algorithm: freq = (clocks in 50 ms) / 50 ms * 1000
	 */
	uint32 start_clocks, end_clocks;
	
	/* wait till beginning of next tick & save it */
	uint32 start_ticks;
	uint32 temp = SDL_GetTicks();
	do
		start_ticks = SDL_GetTicks();
	while(start_ticks == temp);

	__asm
	{
		cpuid
		rdtsc
		mov			[start_clocks], eax
	}

	/* wait at at least 50 ms */
	uint32 end_ticks;
	do
		end_ticks = SDL_GetTicks();
	while(end_ticks < start_ticks+50);

	__asm
	{
		cpuid
		rdtsc
		mov			[end_clocks], eax
	}

	sys_cpu_freq = (end_clocks-start_clocks)/(end_ticks-start_ticks)*1000.f;
	/* NOTE: cpuid/rdtsc overhead is negligible */
}
#endif


/************************************************************************************************
 * init
 ************************************************************************************************/

int sys_init()
{
#ifdef x86
	get_x86_cpu_info();

	_controlfp(_PC_24, MCW_PC);
#endif

	/* success */
	return 0;
}