/*
 * ZIP archive handling / unzip
 *
 * Copyright (c) 2002 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include "precompiled.h"

#include "wposix.h"

#include <SDL/sdl_byteorder.h>
#include <SDL/sdl_endian.h>

#include "unzip.h"


/************************************************************************************************
 * ZIP headers
 ************************************************************************************************/

static const char ecdr_id[] = "PK\5\6";	/* End of Central Directory Header identifier */
static const char cdfh_id[] = "PK\1\2";	/* Central File Header identifier */
static const char lfh_id[]  = "PK\3\4";	/* Local File Header identifier */

/*
 * use #defines instead of structs to avoid padding problems
 * NOTE: byte order conversion is handled automatically
 */

/* Local File Header */
#define LFH_SIZE 30
#define LFH_STRIDE		(SDL_SwapLE16(*(uint16*)&lfh[26])+SDL_SwapLE16(*(uint16*)&lfh[28])+LFH_SIZE)

/* Central Directory File Header */
#define CDFH_SIZE 46
#define CDFH_VER_NEEDED	cdfh[6]		/* version needed to extract */
#define CDFH_FLAGS		cdfh[8]		/* general purpose flags (bit 1: archive is encrypted) */
#define CDFH_CMETHOD	cdfh[10]	/* compression method */
#define CDFH_UCSIZE		SDL_SwapLE32(*(uint32*)&cdfh[24])	/* uncompressed size */
#define CDFH_LFH_OFS	SDL_SwapLE32(*(uint32*)&cdfh[42])	/* local file header offset WRT zip */
#define CDFH_FN_LEN		SDL_SwapLE16(*(uint16*)&cdfh[28])
#define CDFH_STRIDE		(CDFH_FN_LEN+SDL_SwapLE16(*(uint16*)&cdfh[30])+SDL_SwapLE16(*(uint16*)&cdfh[32])+CDFH_SIZE)

/* End of Central Directory Record */
#define ECDR_CD_ENTRIES	SDL_SwapLE16(*(uint16*)&ecdr[10])	/* total central directory entries */
#define ECDR_CD_OFS		SDL_SwapLE32(*(uint32*)&ecdr[16])	/* central directory offset WRT zip */


/************************************************************************************************
 * ZIP handling
 ************************************************************************************************/


/* private data held over open_zip() / close_zip() */
typedef struct _ZIP
{
	uchar* file;		/* pointer to memory mapped zip file */
	long size;			/* file size (needed for munmap()) */

	char* fn;			/* filename */

	uchar* cd;			/* central directory */
	uint16 cd_entries;	/* # entries in the central directory */

	char error[100];	/* buffer for error messages (guarded against overruns) */
}
ZIP;

/*
 * keep track of each zip file we've opened.
 * the application gets a 'zip descriptor' zd (i.e. index into zip_array);
 * each function verifies its zd parameter is in range and valid.
 * zip_cleanup is called at exit to close all open zip files and free zip_array
 */
static ZIP* zip_array;
static uint zip_array_size;
static uint zip_array_entries;


/* close all open zip files and free zip_array (at exit) */
static void cleanup(void)
{
	int zd;
	for(zd = 0; (unsigned)zd < zip_array_size; zd++)
		close_zip(zd);
	free(zip_array);
}


/* close the zip file zd */
void close_zip(const int zd)
{
	ZIP* zip;

	if((unsigned)zd >= zip_array_size || !zip_array[zd].file)
		return;

	zip = &zip_array[zd];

	free(zip->fn);
	munmap(zip->file, zip->size);

	/* mark this entry as unused */
	zip->file = NULL;
}


/*
 * open a zip file and return its 'zip descriptor'
 * (a small, non-negative integer for use in unzip(), close_zip(), zip_error_str())
 *
 * returns -1 if the file couldn't be memory mapped, or isn't a valid ZIP file
 *
 * does not update the error string, since we don't allocate a zip descriptor
 * until we're sure the function will succeed (which avoids messy cleanup)
 *
 *
 * memory maps the file, finds the end of central directory record (ecdr),
 * and saves the location of the central directory
 */
int open_zip(const char* fn)
{
	/* first call: register cleanup atexit routine */
	if(!zip_array)
		atexit(cleanup);

	/* open zip file */
	struct stat stat_buf;
	const int fd = open(fn, O_RDONLY);
	if(fd < 0 || stat(fn, &stat_buf))
		return -1;

	/* map zip file */
	const long size = stat_buf.st_size;
	uchar* file = (uchar*)mmap(0, size, PROT_READ, MAP_PRIVATE, fd, 0);
	close(fd);
	if(file == MAP_FAILED)
		return -1;

	/*
	 * find end of central dir record
	 * by scanning last 66000 bytes of file for ecdr_id magic
	 * (zip comment <= 65535 bytes, sizeof(ECDR) = 22,
	 *	add some for safety)
	 * if the zip file is < 66000 bytes, scan the whole file
	 */
	uchar* ecdr = max(file, file+size-66000);	/* scan from start of file if < 66kb */
	uint bytes_left = min(66000, size);		/* scan whole file if < 66 kb */

	while(bytes_left)
	{
		/* look for ecdr id */
		if(*(uint32*)ecdr == *(uint32*)&ecdr_id)	/* found the ecdr? */
			break;									/* yes, quit */

		/* check next 4 bytes (non aligned!!) */
		ecdr++; bytes_left--;
	}

	/* reached EOF and still haven't found the ecdr identifier */
	if(bytes_left == 0)
		return -1;

	/* alloc entry */
	if(zip_array_entries >= zip_array_size)
	{
		zip_array_size += 4;
		zip_array = (ZIP*)realloc(zip_array, zip_array_size*sizeof(ZIP));

		/* mark new entries as free (i.e. ZIP.file == NULL) */
		memset(&zip_array[zip_array_entries], 0, 4*sizeof(ZIP));
	}
	const int zd = zip_array_entries++;
	ZIP* zip = &zip_array[zd];

	/* write ZIP entry fields */
	zip->file = file;
	zip->size = size;
	zip->fn = (char*)malloc(strlen(fn)+1);
	strcpy(zip->fn, fn);
	zip->cd = file+ECDR_CD_OFS;			/* save address of the central dir */
	zip->cd_entries = ECDR_CD_ENTRIES;	/* ... and # entries for use in unzip() */

	return zd;
}


static uchar* in;		/* pointer to compressed data in the memory mapped zip file */
static uchar* out;		/* output buffer (size == lfh->ucsize) */

static int inflate();


/*
 * extracts fn from the zip file zd;
 * returns a pointer to the data and (if size != NULL) its length in size
 *
 * returns NULL on error.
 *
 * NOTES:
 *  - it is the application's responsibility to deallocate
 *    memory returned by this function via free()
 *
 *
 * scans through all file names in the central directory;
 * if fn is found, its format (must be unencrypted, < Zip 2.0 deflated or stored)
 * is checked; finally, in / out pointers are initialized and the file is decompressed
 */
void* unzip(const int zd, const char* fn, int* size)
{
	/* *size = 0 if unzip() fails */
	if(size)
		*size = 0;

	/* make sure zd is valid */
	if((unsigned)zd >= zip_array_size || !zip_array[zd].file)
		return NULL;
	ZIP* zip = &zip_array[zd];

	/*
	 * things that can go wrong:
	 * 1) zip file is damaged
	 * 2) file not found in archive
	 * 3) unsupported zip format (i.e. not < Zip 2.0 deflated or stored)
	 *
	 * simplify error reporting by assuming error 1
	 * and only changing the error string for errors 2 and 3
	 */
	sprintf(zip->error, "Problem loading \"%.20s\" from \"%.20s\": archive is damaged", fn, zip->fn);

	/* find fn's file header in the central directory */
	const uchar *cdfh = zip->cd, *lfh;
	const int fn_len = strlen(fn);

	uint16 cd_entries_left = zip->cd_entries;
	while(cd_entries_left)
	{
		/* verify cdfh id */
		if(*(uint32*)cdfh != *(uint32*)&cdfh_id)	/* cdfh valid? */
			return NULL;							/* no - quit with "damaged" error string */

		lfh = zip->file+CDFH_LFH_OFS;

		/*
		 * compare filename
		 * NOTE: filename is stored after file header and is not 0-terminated!
		 */
		const char* temp_fn = (char*)cdfh+CDFH_SIZE;
		if(fn_len == CDFH_FN_LEN &&			/* are file names the same length? */
		   !strncmp(fn, temp_fn, fn_len))	/* and equal? */
			break;							/* yes - exit loop */

		/* advance to next cdfh */
		cdfh += CDFH_STRIDE;
		cd_entries_left--;
	}

	if(cd_entries_left == 0)
	{
		sprintf(zip->error, "Problem extracting \"%.20s\" from \"%.20s\": file not found in archive", fn, &zip->fn);
		return NULL;
	}

	/* read cdfh fields; make sure we can extract the file */
	const int ucsize = CDFH_UCSIZE;
	if(CDFH_FLAGS & 1 ||		/* encrypted */
	   CDFH_VER_NEEDED > 20 ||	/* > ZIP 2.0 */
	   CDFH_CMETHOD & (~8))		/* neither deflated (8) nor stored (0) */
	{
		sprintf(zip->error, "Problem extracting \"%.20s\" from \"%.20s\": invalid format (only Zip 2.0 deflated/stored, non-encrypted archives are supported)", fn, &zip->fn);
		return NULL;
	}

	/* verify lfh id */
	if(*(uint32*)lfh != *(uint32*)&lfh_id)	/* lfh valid? */
		return NULL;						/* no - quit with "damaged" error string */

	/* set source/dest pointers */
	in = (uchar*)lfh+LFH_STRIDE;
	out = (uchar*)malloc(ucsize);

	/* file is stored */
	if(CDFH_CMETHOD == 0)
		memcpy(out, in, ucsize);	/* have to copy because in is mapped read only */
	/* file is deflated */
	else
		if(!inflate())
		{
			free(out);
			return NULL;
		}

	if(size)
		*size = ucsize;

	return (void*)out;
}


/*
 * returns a string explaining the last error.
 * valid only after unzip() returns NULL and before
 * the next unzip() call with the same zip descriptor
 */
const char* last_zip_error(const int zd)
{
	/* make sure zd is valid */
	if((unsigned)zd >= zip_array_size || !zip_array[zd].file)
		return NULL;

	return zip_array[zd].error;
}


/************************************************************************************************
 * decompression (based on Mark Adler's public domain inflate.c)
 ************************************************************************************************/

/*
	Inflate deflated (PKZIP's method 8 compressed) data.  The compression
	method searches for as much of the current string of bytes (up to a
	length of 258) in the previous 32K bytes.  If it doesn't find any
	matches (of at least length 3), it codes the next byte.  Otherwise, it
	codes the length of the matched string and its distance backwards from
	the current position.  There is a single Huffman code that codes both
	single bytes (called "literals") and match lengths.  A second Huffman
	code codes the distance information, which follows a length code.  Each
	length or distance code actually represents a base value and a number
	of "extra" (sometimes zero) bits to get to add to the base value.  At
	the end of each deflated block is a special end-of-block (EOB) literal/
	length code.  The decoding process is basically: get a literal/length
	code; if EOB then done; if a literal, emit the decoded byte; if a
	length then get the distance and emit the referred-to bytes from the
	sliding window of previously emitted data.

	There are (currently) three kinds of inflate blocks: stored, fixed, and
	dynamic.  The compressor outputs a chunk of data at a time and decides
	which method to use on a chunk-by-chunk basis.  A chunk might typically
	be 32K to 64K, uncompressed.  If the chunk is uncompressible, then the
	"stored" method is used.  In this case, the bytes are simply stored as
	is, eight bits per byte, with none of the above coding.  The bytes are
	preceded by a count, since there is no longer an EOB code.

	If the data is compressible, then either the fixed or dynamic methods
	are used.  In the dynamic method, the compressed data is preceded by
	an encoding of the literal/length and distance Huffman codes that are
	to be used to decode this block.  The representation is itself Huffman
	coded, and so is preceded by a description of that code.  These code
	descriptions take up a little space, and so for small blocks, there is
	a predefined set of codes, called the fixed codes.  The fixed method is
	used if the block ends up smaller that way (usually for quite small
	chunks); otherwise the dynamic method is used.  In the latter case, the
	codes are customized to the probabilities in the current block and so
	can code it much better than the pre-determined fixed codes can.
 
	The Huffman codes themselves are decoded using a mutli-level table
	lookup, in order to maximize the speed of decoding plus the speed of
	building the decoding tables.  See the comments below that precede the
	lbits and dbits tuning parameters.

	Notes beyond the 1.93a appnote.txt:

	 1. Distance pointers never point before the beginning of the output stream.
	 2. Distance pointers can point back across blocks, up to 32k away.
	 3. There is an implied maximum of 7 bits for the bit length table and
	    15 bits for the actual data.
	 4. If only one code exists, then it is encoded using one bit.
	    (Zero would be more efficient, but perhaps a little confusing.)
	    If two codes exist, they are coded using one bit each (0 and 1).
	 5. There is no way of sending zero distance codes--a dummy must be
	    sent if there are none.  (History: a pre 2.0 version of PKZIP would
	    store blocks with no distance codes, but this was discovered to be
	    too harsh a criterion.)  Valid only for 1.93a.  2.04c does allow
	    zero distance codes, which is sent as one code of zero bits in length.
	 6. There are up to 286 literal/length codes.  Code 256 represents the
	    end-of-block.  Note however that the static length tree defines
	    288 codes just to fill out the Huffman codes.  Codes 286 and 287
	    cannot be used though, since there is no length base or extra bits
	    defined for them.  Similarily, there are up to 30 distance codes.
	    However, static trees define 32 codes (all 5 bits) to fill out the
	    Huffman codes, but the last two had better not show up in the data.
	 7. Unzip can check dynamic Huffman blocks for complete code sets.
	    The exception is that a single code would not be complete (see #4).
	 8. The five bits following the block type is really the number of
	    literal codes sent minus 257.
	 9. Length codes 8,16,16 are interpreted as 13 length codes of 8 bits
	    (1+6+6).  Therefore, to output three times the length, you output
	    three codes (1+1+1), whereas to output four times the same length,
	    you only need two codes (1+3).  Hmm.
	10. In the tree reconstruction algorithm, Code = Code + Increment
	    only if BitLength(i) is not zero.  (Pretty obvious.)
	11. Correction: 4 Bits: # of Bit Length codes - 4     (4 - 19)
	12. Note: length code 284 can represent 227-258, but length code 285
	    really is 258.  The last length deserves its own, short code
	    since it gets used a lot in very redundant files.  The length
	    258 is special since 258 - 3 (the min match length) is 255.
	13. The literal/length and distance code bit lengths are read as a
	    single stream of lengths.  It is possible (and advantageous) for
	    a repeat code (16, 17, or 18) to go across the boundary between
	    the two sets of lengths.
*/

#define WSIZE 0x80000000	/* window size = 2^x, x > 15 */


/************************************************************************************************
 * Huffman tree routines
 ************************************************************************************************/

/*
 * huffman code lookup table entry
 * valid extra bits (e) are [13:0]:
 *   e == 15 is EOB (end of block)
 *   e == 16 means that v is a literal
 *   16<e<32 means that v is a pointer to the next table,
 *     which codes e - 16 bits
 *   e == 99 indicates an unused code. If a code with e == 99 is looked up,
 *     this implies an error in the data.
 */
typedef struct _HUFT
{
	uchar e;				/* number of extra bits or operation */
	uchar b;				/* number of bits in this code or subcode */
	union
	{
		uint16 n;			/* literal, length base, or distance base */
		struct _HUFT* t;	/* pointer to next level of table */
	}
	v;
}
HUFT;


/*
 * the deflate algorithm uses a sliding 32k byte window on the uncompressed
 * stream to find repeated byte strings.
 * here, the 'window' contains the whole file (out)
 */
static uint w;	/* position in out */


/*
 * deflate tables from pkzip's appnote.txt
 */

/* order of the bit length code lengths */
static uint border[] = { 16, 17, 18, 0, 8, 7, 9, 6, 10, 5, 11, 4, 12, 3, 13, 2, 14, 1, 15};

/*
 * copy lengths for literal codes 257..285
 * NOTE: see #13 above about the 258 in this list
 */
static uint16 cplens[] =
{
	3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 15, 17, 19, 23, 27, 31,
	35, 43, 51, 59, 67, 83, 99, 115, 131, 163, 195, 227, 258, 0, 0
};

/* extra bits for literal codes 257..285 */
static uint16 cplext[] =
{
	0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 2, 2, 2, 2,
	3, 3, 3, 3, 4, 4, 4, 4, 5, 5, 5, 5, 0, 99, 99	/* 99 is invalid */
};

/* copy offsets for distance codes 0..29 */
static uint16 cpdist[] =
{
	1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193,
	257, 385, 513, 769, 1025, 1537, 2049, 3073, 4097, 6145,
	8193, 12289, 16385, 24577
};

/* extra bits for distance codes */
static uint16 cpdext[] =
{
	0, 0, 0, 0, 1, 1, 2, 2, 3, 3, 4, 4, 5, 5, 6, 6,
	7, 7, 8, 8, 9, 9, 10, 10, 11, 11, 12, 12, 13, 13
};

/* mask[n] = mask for lower n bits */
static uint16 mask[] =
{
		0, 1, 3, 7, 15, 31, 63, 127, 255, 511,
		1023, 2047, 4095, 8191, 16383, 31767, 65535
};


/*
 * input bit buffer
 */

static uint32 b;	/* bit buffer */
static uint k;	/* bits in bit buffer */

/*
 * buffer access macros
 *
 * NOTE: local variables b, k must be copied from bb, bk before use
 *
 * In order to not ask for more bits than there are in the compressed
 * stream, the Huffman tables are constructed to only ask for just
 * enough bits to make up the end-of-block code (value 256).  Then no
 * bytes need to be "returned" to the buffer at the end of the last
 * block.  See huft_build().
 */

/* make sure the buffer contains n bits */
#define GETBITS(n) {while(k < (n)) {b |= ((uint32)(*in++)) << k; k += 8;}}

/* remove n bits from the buffer (used after GETBITS(n)) */
#define REMOVEBITS(n) {b >>= (n); k -= (n);}


/*
	 Huffman code decoding is performed using a multi-level table lookup.
	 The fastest way to decode is to simply build a lookup table whose
	 size is determined by the longest code.	However, the time it takes
	 to build this table can also be a factor if the data being decoded
	 is not very long.	The most common codes are necessarily the
	 shortest codes, so those codes dominate the decoding time, and hence
	 the speed.	The idea is you can have a shorter table that decodes the
	 shorter, more probable codes, and then point to subsidiary tables for
	 the longer codes.	The time it costs to decode the longer codes is
	 then traded against the time it takes to make longer tables.

	 This results of this trade are in the variables lbits and dbits
	 below.	lbits is the number of bits the first level table for literal/
	 length codes can decode in one step, and dbits is the same thing for
	 the distance codes.	Subsequent tables are also less than or equal to
	 those sizes.	These values may be adjusted either when all of the
	 codes are shorter than that, in which case the longest code length in
	 bits is used, or when the shortest code is *longer* than the requested
	 table size, in which case the length of the shortest code in bits is
	 used.

	 There are two different values for the two tables, since they code a
	 different number of possibilities each.	The literal/length table
	 codes 286 possible values, or in a flat code, a little over eight
	 bits.	The distance table codes 30 possible values, or a little less
	 than five bits, flat.	The optimum values for speed end up being
	 about one bit more than those, so lbits is 8+1 and dbits is 5+1.
	 The optimum values may differ though from machine to machine, and
	 possibly even between compilers.
*/

static int lbits = 9;	/* bits in base literal/length lookup table */
static int dbits = 6;	/* bits in base distance lookup table */


/* If BMAX needs to be larger than 16, then h and x[] should be uint32. */
#define BMAX 16			 /* maximum bit length of any code (16 for explode) */
#define N_MAX 288		 /* maximum number of codes in any set */


static uint hufts;		 /* track memory usage */

static void huft_free(HUFT* t);

/*
 * Given a list of code lengths and a maximum table size, make a set of
 * tables to decode that set of codes.	Return zero on success, one if
 * the given code set is incomplete (the tables are still built in this
 * case), two if the input is invalid (all zero length codes or an
 * oversubscribed set of lengths), and three if not enough memory.
 * The code with value 256 is special, and the tables are constructed
 * so that no bits beyond that code are fetched when that code is decoded.
 *
 * params:
 *   uint* b:   code lengths in bits (assumed <= BMAX)
 *   uint n:    number of codes (assumed <= N_MAX)
 *   uint s:    number of simple-valued codes (0..s-1)
 *   uint16* d: list of base values for non-simple codes
 *   uint16* e: list of extra bits for non-simple codes
 *   HUFT** t:  result: starting table
 *   int* m:    maximum lookup bits, returns actual
 */
static int huft_build(uint* b, uint n, uint s, uint16* d, uint16* e, HUFT** t, int* m)
{
	uint a;					/* counter for codes of length k */
	uint c[BMAX+1];			/* bit length count table */
	uint el;				/* length of EOB code (value 256) */
	uint f;					/* i repeats in table every f entries */
	int g;					/* maximum code length */
	int h;					/* table level */
	uint i, j;				/* counter, current code / counter */
	int k;					/* number of bits in current code */
	int lx[BMAX+1];			/* memory for l[-1..BMAX-1] */
	int *l = lx+1;			/* stack of bits per table */
	uint *p;				/* pointer into c[], b[], or v[] */
	HUFT *q;				/* points to current table */
	HUFT r;					/* table entry for structure assignment */
	HUFT *u[BMAX];			/* table stack */
	static uint v[N_MAX];	/* values in order of bit length */
	int w;					/* bits before this table == (l * h) */
	uint x[BMAX+1];			/* bit offsets, then code stack */
	uint *xp;				/* pointer into x */
	int y;					/* number of dummy codes added */
	uint z;					/* number of entries in current table */


	/* Generate counts for each bit length */
	el = n > 256 ? b[256] : BMAX; /* set length of EOB code, if any */
	memset((char*)c, 0, sizeof(c));
	p = b;	i = n;
	do
		c[*p++]++;	/* assume all entries <= BMAX */
	while (--i);
	if(c[0] == n)	/* null input--all zero length codes */
	{
		*t = NULL;
		*m = 0;
		return 1;
	}


	/* Find minimum and maximum length; bound *m by those */
	for (j = 1; j <= BMAX; j++)
		if(c[j])
			break;
	k = j;				/* minimum code length */
	if((uint)*m < j)
		*m = j;
	for (i = BMAX; i; i--)
		if(c[i])
			break;
	g = i;				/* maximum code length */
	if((uint)*m > i)
		*m = i;


	/* Adjust last length count to fill out codes, if needed */
	for (y = 1 << j; j < i; j++, y <<= 1)
		if((y -= c[j]) < 0)
			return 0;	/* bad input: more codes than bits */
	if((y -= c[i]) < 0)
		return 0;
	c[i] += y;


	/* Generate starting offsets into the value table for each length */
	x[1] = j = 0;
	p = c + 1;	xp = x + 2;
	while (--i)			 /* note that i == g from above */
		*xp++ = (j += *p++);


	/* Make a table of values in order of bit lengths */
	p = b;	i = 0;
	do
		if((j = *p++) != 0)
			v[x[j]++] = i;
	while (++i < n);


	/* Generate the Huffman codes and for each, make the table entries */
	x[0] = i = 0;	/* first Huffman code is zero */
	p = v;			/* grab values in bit order */
	h = -1;			/* no tables yet -> level -1 */
	w = l[-1] = 0;	/* no bits decoded yet */


	/* go through the bit lengths (k already is bits in shortest code) */
	for (; k <= g; k++)
	{
		a = c[k];
		while (a--)
		{
			/* here i is the Huffman code of length k bits for value *p */
			/* make tables up to required level */
			while (k > w + l[h])
			{
				w += l[h++];	/* add bits already decoded */

				/* compute minimum size table less than or equal to *m bits */
				z = (z = g - w) > (uint)*m ? *m : z; /* upper limit */
				if((f = 1 << (j = k - w)) > a + 1)	 /* try a k-w bit table */
				{									 /* too few codes for k-w bit table */
					f -= a + 1;						 /* deduct codes from patterns left */
					xp = c + k;
					while (++j < z)			/* try smaller tables up to z bits */
					{
						if((f <<= 1) <= *++xp)
							break;			/* enough codes to use up j bits */
						f -= *xp;			/* else deduct codes from patterns */
					}
				}
				if((uint)w + j > el && (uint)w < el)
					j = el - w;				/* make EOB code end at table */
				z = 1 << j;					/* table entries for j-bit table */
				l[h] = j;					/* set table size in stack */

				/* allocate and link in new table */
				q = (HUFT*)malloc((z+1)*sizeof(HUFT));
				hufts += z+1;				/* track memory usage */
				*t = q + 1;					/* link to list for huft_free() */
				*(t = &(q->v.t)) = NULL;
				u[h] = ++q;					/* table starts after link */

				/* connect to last table, if there is one */
				if(h)
				{
					x[h] = i;				/* save pattern for backing up */
					r.b = (uchar)l[h-1];	/* bits to dump before this table */
					r.e = (uchar)(16 + j);	/* bits in this table */
					r.v.t = q;				/* pointer to this table */
					j = (i & ((1 << w) - 1)) >> (w - l[h-1]);
					u[h-1][j] = r;			/* connect to last table */
				}
			}

			/* set up table entry in r */
			r.b = (uchar)(k - w);
			if(p >= v + n)
				r.e = 99;					/* out of values--invalid code */
			else if(*p < s)
			{
				r.e = (uchar)(*p < 256? 16 : 15);	/* 256 is end-of-block code */
				r.v.n = (uint16)*p++;				/* simple code is just the value */
			}
			else
			{
				r.e = (uchar)e[*p - s];	/* non-simple--look up in lists */
				r.v.n = d[*p++ - s];
			}

			/* fill code-like entries with r */
			f = 1 << (k - w);
			for (j = i >> w; j < z; j += f)
				q[j] = r;

			/* backwards increment the k-bit code i */
			for (j = 1 << (k - 1); i & j; j >>= 1)
				i ^= j;
			i ^= j;

			/* backup over finished tables */
			while ((i & ((1 << w) - 1)) != x[h])
				w -= l[--h];						/* don't need to update q */
		}
	}


	/* pass actual size of base table */
	*m = l[0];


	/* return 0 if we were given an incomplete table */
	return y == 0 || g == 1;
}


/*
 * free all tables allocated by huft_build()
 * (tables are linked by pointers in their first entry)
 */
static void huft_free(HUFT* t)
{
	HUFT *p, *q;

	/* go through linked list, freeing from the malloc'ed (t[-1]) address. */
	p = t;
	while (p != NULL)
	{
		q = (--p)->v.t;
		free(p);
		p = q;
	} 
}


/************************************************************************************************
 * inflation routines
 ************************************************************************************************/

/*
 * inflate (decompress) the codes in a deflated (compressed) block;
 * return 1 on success.
 *
 * params:
 *   HUFT* tl:
 *   HUFT* td: literal/length and distance decoder tables
 *   uint bl:
 *   uint bd: number of bits decoded by tl[] and td[]
 */
static int inflate_codes(HUFT* tl, HUFT* td, uint bl, uint bd)
{
	uint e;			/* table entry flag/number of extra bits */
	uint n, d;		/* length and index for copy */
	HUFT *t;		/* pointer to table entry */
	uint ml, md;	/* masks for bl and bd bits */

	/* inflate the coded data */
	ml = mask[bl];	/* precompute masks for speed */
	md = mask[bd];
	while(1)		/* do until end of block */
	{
		GETBITS(bl)																	/* 13000 */
		if((e = (t = tl + ((uint)b & ml))->e) > 16)									/* 20000 */
			do
			{
				if(e == 99)
					return 0;
				REMOVEBITS(t->b)
				e -= 16;
				GETBITS(e)
			}
			while ((e = (t = t->v.t + ((uint)b & mask[e]))->e) > 16);
		REMOVEBITS(t->b)													/* 5 */
		/* it's a literal */
		if(e == 16)
			out[w++] = (uchar)t->v.n;										/* 3 */
		/* it's an EOB or a length */
		else
		{
			/* exit if end of block */
			if(e == 15)
				break;

			/* get length of block to copy */
			GETBITS(e)
			n = t->v.n + ((uint)b & mask[e]);
			REMOVEBITS(e);

			/* decode distance of block to copy */
			GETBITS((uint)bd)
			if((e = (t = td + ((uint)b & md))->e) > 16)						/* 5 */
				do
				{
					if(e == 99)
						return 0;
					REMOVEBITS(t->b)
					e -= 16;
					GETBITS(e)
				}
				while ((e = (t = t->v.t + ((uint)b & mask[e]))->e) > 16);
			REMOVEBITS(t->b)
			GETBITS(e)
			d = w - t->v.n - ((uint)b & mask[e]);
			REMOVEBITS(e)

			/* do the copy */
			do
			{
				e = WSIZE - max(d,w);
				e = e > n? n : e;
				n -= e;

				if(w - d >= e)		/* assume unsigned compare */
				{
					memcpy(out+w, out+d, e);								/* 9 */
					w += e;
					d += e;
				}
				/* regions overlap, copy slow */
				else
					do
						out[w++] = out[d++];
					while(--e);
			}
			while (n);
		}
	}

	return 1;
}


/* "decompress" an inflated type 0 (stored) block */
static int inflate_stored()
{
	uint n;			/* number of bytes in block */

	/* skip to next byte boundary */
	n = k&7;
	REMOVEBITS(n);

	/* get the block length and its complement */
	GETBITS(16)
	n = ((uint)b & 0xffff);
	REMOVEBITS(16)
	GETBITS(16)
	if(n != (uint)((~b) & 0xffff))
		return 0;	/* error in compressed data */
	REMOVEBITS(16)

	/* copy stored block */
	memcpy(&out[w], in, n);
	w += n;

	return 1;
}


/* globals for literal tables (built once) */
static HUFT* fixed_tl = NULL;
static HUFT* fixed_td;
static int fixed_bl, fixed_bd;


/*
 * decompress an inflated type 1 (fixed Huffman codes) block.
 *
 * TODO: replace this with a custom decoder,
 * or at least precompute the Huffman tables
 */
static int inflate_fixed()
{
	/* if first time, set up tables for fixed blocks */
	if(fixed_tl == NULL)
	{
		int i;
		static unsigned l[288];		/* length list for huft_build */

		/* literal table */
		for (i = 0; i < 144; i++)
			l[i] = 8;
		for (; i < 256; i++)
			l[i] = 9;
		for (; i < 280; i++)
			l[i] = 7;
		for (; i < 288; i++)		/* make a complete, but wrong code set */
			l[i] = 8;
		fixed_bl = 7;
		if((i = huft_build(l, 288, 257, cplens, cplext, &fixed_tl, &fixed_bl)) != 0)
		{
			fixed_tl = NULL;
			return i;
		}

		/* distance table */
		for (i = 0; i < 30; i++)	/* make an incomplete code set */
			l[i] = 5;
		fixed_bd = 5;
		if(!huft_build(l, 30, 0, cpdist, cpdext, &fixed_td, &fixed_bd))
		{
			huft_free(fixed_tl);
			fixed_tl = NULL;
			return 0;
		}
	}


	/* decompress until an end-of-block code */
	return inflate_codes(fixed_tl, fixed_td, fixed_bl, fixed_bd) == 0;
}


/* decompress an inflated type 2 (dynamic Huffman codes) block */
static int inflate_dynamic()
{
	int i;
	uint j;
	uint l;					/* last length */
	uint m;					/* mask for bit lengths table */
	uint n;					/* number of lengths to get */
	HUFT* tl;				/* literal/length code table */
	HUFT* td;				/* distance code table */
	int bl, bd;				/* lookup bits for tl/td */
	uint nb;				/* number of bit length codes */
	uint nl;				/* number of literal/length codes */
	uint nd;				/* number of distance codes */
	static uint ll[286+30]; /* literal/length and distance code lengths */


	/* read in table lengths */
	GETBITS(5)
	nl = 257 + ((uint)b & 0x1f);		/* number of literal/length codes */
	REMOVEBITS(5)
	GETBITS(5)
	nd = 1 + ((uint)b & 0x1f);			/* number of distance codes */
	REMOVEBITS(5)
	GETBITS(4)
	nb = 4 + ((uint)b & 0xf);			/* number of bit length codes */
	REMOVEBITS(4)
	if(nl > 286 || nd > 30)
		return 0;						/* bad lengths */


	/* read in bit-length-code lengths */
	for (j = 0; j < nb; j++)
	{
		GETBITS(3)
		ll[border[j]] = (uint)b & 7;
		REMOVEBITS(3)
	}
	for (; j < 19; j++)
		ll[border[j]] = 0;


	/* build decoding table for trees--single level, 7 bit lookup */
	bl = 7;
	if(!huft_build(ll, 19, 19, NULL, NULL, &tl, &bl))
	{
		huft_free(tl);
		return 0;	/* incomplete code set */
	}


	/* read in literal and distance code lengths */
	n = nl + nd;
	m = mask[bl];
	i = l = 0;
	while ((uint)i < n)
	{
		GETBITS((uint)bl)
		j = (td = tl + ((uint)b & m))->b;
		REMOVEBITS(j)
		j = td->v.n;
		if(j < 16)						/* length of code in bits (0..15) */
			ll[i++] = l = j;			/* save last length in l */
		else if(j == 16)				/* repeat last length 3 to 6 times */
		{
			GETBITS(2)
			j = 3 + ((uint)b & 3);
			REMOVEBITS(2)
			if((uint)i + j > n)
				return 0;
			while (j--)
				ll[i++] = l;
		}
		else if(j == 17)				/* 3 to 10 zero length codes */
		{
			GETBITS(3)
			j = 3 + ((uint)b & 7);
			REMOVEBITS(3)
			if((uint)i + j > n)
				return 0;
			while (j--)
				ll[i++] = 0;
			l = 0;
		}
		else							/* j == 18: 11 to 138 zero length codes */
		{
			GETBITS(7)
			j = 11 + ((uint)b & 0x7f);
			REMOVEBITS(7)
			if((uint)i + j > n)
				return 0;
			while (j--)
				ll[i++] = 0;
			l = 0;
		}
	}

	huft_free(tl);	/* free decoding table for trees */


	/* build the decoding tables for literal/length and distance codes */
	bl = lbits;
	if(!huft_build(ll, nl, 257, cplens, cplext, &tl, &bl))
	{
		huft_free(tl);
		return 0;
	}
	bd = dbits;
	if(!huft_build(ll + nl, nd, 0, cpdist, cpdext, &td, &bd))
	{
		huft_free(tl);
		huft_free(td);
		return 0;
	}

	/* decompress until an end-of-block code */
	if(!inflate_codes(tl, td, bl, bd))
		return 0;


	/* free the decoding tables */
	huft_free(tl);
	huft_free(td);
	return 1;
}


/*
 * decompress a deflated file
 * IN: in -> file to decomress
 */
static int inflate()
{
	uint is_last_block;	/* is this block the last one? */
	uint t;				/* block type */

	/* initialize window, bit buffer */
	w = k = b = 0;

	/* decompress blocks */
	while(1)
	{
		hufts = 0;

		/* read in last block bit */
		GETBITS(1)
		is_last_block = (uint)b&1;
		REMOVEBITS(1)

		/* read in block type */
		GETBITS(2)
		t = (uint)b&3;
		REMOVEBITS(2)

		/* inflate that block type */
		if(t == 0)
			if(!inflate_stored())
				return 0;
		if(t == 1)
			if(!inflate_fixed())
				return 0;
		if(t == 2)
			if(!inflate_dynamic())
				return 0;

		if(is_last_block)
			return 1;
	}
}