/*
 * misc util functions
 *
 * Copyright (c) 2002 Jan Wassenberg
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * Contact info:
 *   Jan.Wassenberg@stud.uni-karlsruhe.de
 *   http://www.stud.uni-karlsruhe.de/~urkt/
 */

#include "precompiled.h"

#include "wsdl.h"

#include "system.h"


int log2(const int n)
{
	int i = 1, j = 0;
	for(; i != n; i += i, j++);
	return j;
}



#define H 10	/* history buffer size */

int fps = 0;

void calc_fps()
{
	/* calculate the current fps value (= 1/(time elapsed since last call)) */
	static uint32 start_time = 0;

#ifdef x86
	uint32 end_time;
	__asm
	{
		cpuid
		rdtsc
		mov			[end_time], eax
	}
	const float cur_fps = sys_cpu_freq/(end_time-start_time);	/* [frames/s] = [clocks/s]/[clocks/frame] */
#else
	const uint32 end_time = SDL_GetTicks();
	const float cur_fps = 1000.f/(end_time-start_time);		/* [frames/s] = [1000 ms/s]/[ms / frame] */
#endif

	static float new_fps = 0;	/* average of last x frames' cur_fps;
								   used to update fps */
	static float fps_hist[100];	/* last x frames' cur_fps/x (ringbuffer for speed) */
	static int head = 0;		/* oldest value in fps_hist */

	float h1 = fps_hist[(head-1)%H];	/* last frame's cur_fps */
	float h2 = fps_hist[(head-2)%H];	/* 2nd most recent frame's cur_fps */

	/*
	 * calculate fps activity over 3 frames (used below to prevent fluctuation)
	 * -1: decreasing, +1: increasing, 0: neither or fluctuating
	 */
	int trend = 0;
	if(h2 > h1 && h1 > cur_fps)
		trend = -1;
	else if(h2 < h1 && h1 < cur_fps)
		trend = 1;

	/* filter out dips in fps (probably page faults) */
	static int bad = 0;	/* bad > 0 <==> last value was skipped */
	if(fabs(h1-cur_fps) > .05f*h1)
	{
		/*
		 * if this is the first time a bad value is encountered
		 * (bad == 0), don't update fps_hist/fps; otherwise set bad > 0
		 */
		if(!bad++)
			goto end;	/* still have to save start time for next frame */
	}
	else
		bad = 0;

	/*
	 * remove oldest cur_fps value in fps_hist from the average
	 * and add cur_fps; also insert cur_fps in fps_hist
	 */
	new_fps -=  fps_hist[head];
	new_fps += (fps_hist[head] = cur_fps/H);
	head = (head+1)%H;

	/* update fps counter if update threshold is exceeded */
	if((trend > 0 && (new_fps > fps || new_fps-fps < -4.f)) ||
	   (trend < 0 && (new_fps < fps || new_fps-fps > 4.f)) ||
	   (fabs(fps-new_fps) > 5.f))
		fps = (int)new_fps;

end:
	/* save start time for next frame */
#ifdef x86
	__asm
	{
		cpuid
		rdtsc
		mov			[start_time], eax
	}
#else
	start_time = SDL_GetTicks();
#endif
}

